@@ -1,4 +1,4 @@
-# Copyright (c) 2005-2007 Canonical
+# Copyright (c) 2005-2009 Canonical Ltd
 #
 # AUTHOR:
 # Michael Vogt <mvo@ubuntu.com>
@@ -22,7 +22,7 @@
 
 import warnings
 from warnings import warn
-warnings.filterwarnings("ignore", "apt API not stable yet", FutureWarning)
+warnings.filterwarnings("ignore", "Accessed deprecated property", DeprecationWarning)
 import apt_inst, apt_pkg
 import apt
 import sys
@@ -30,8 +30,13 @@ import os
 from gettext import gettext as _
 from Cache import Cache
 
+from debian_bundle.debfile import DebFile
+
+import gzip
+from StringIO import StringIO
+
 class DebPackage(object):
-    debug = 0
+    debug = int(os.environ.get("GDEBI_DEBUG_LEVEL") or 0)
 
     def __init__(self, cache, file=None):
         cache.clear()
@@ -68,11 +73,22 @@ class DebPackage(object):
                         if pkg.isInstalled:
                             return True
                 continue
-
+            # check real dependency
             inst = self._cache[depname]
             instver = inst.installedVersion
             if instver != None and apt_pkg.CheckDep(instver,oper,ver) == True:
                 return True
+
+            # if no real dependency is installed, check if there is
+            # a package installed that provides this dependency
+            # (e.g. scrollkeeper dependecies are provided by rarian-compat)
+            # but only do that if there is no version required in the 
+            # dependency (we do not supprot versionized dependencies)
+            if not oper:
+                for ppkg in self._cache.getProvidersFor(depname):
+                    if ppkg.isInstalled:
+                        self._dbg(3, "found installed '%s' that provides '%s'" % (ppkg.name, depname))
+                        return True
         return False
             
 
@@ -116,9 +132,11 @@ class DebPackage(object):
         or_str = ""
         for dep in or_group:
             or_str += dep[0]
+            if ver and oper:
+                or_str += " (%s %s)" % (dep[2], dep[1])
             if dep != or_group[len(or_group)-1]:
                 or_str += "|"
-        self._failureString += _("Dependency is not satisfiable: %s\n" % or_str)
+        self._failureString += _("Dependency is not satisfiable: %s\n") % or_str
         return False
 
     def _checkSinglePkgConflict(self, pkgname, ver, oper):
@@ -126,6 +144,7 @@ class DebPackage(object):
             pkg """
         # FIXME: deal with conflicts against its own provides
         #        (e.g. Provides: ftp-server, Conflicts: ftp-server)
+        self._dbg(3, "_checkSinglePkgConflict() pkg='%s' ver='%s' oper='%s'" % (pkgname, ver, oper))
         pkgver = None
         cand = self._cache[pkgname]
         if cand.isInstalled:
@@ -136,8 +155,9 @@ class DebPackage(object):
         #print "ver: %s" % ver
         #print "pkgver: %s " % pkgver
         #print "oper: %s " % oper
-        if pkgver and apt_pkg.CheckDep(pkgver,oper,ver):
-            self._failureString += _("Conflicts with the installed package '%s'" % cand.name)
+        if (pkgver and apt_pkg.CheckDep(pkgver,oper,ver) and 
+            not self.replacesRealPkg(pkgname, oper, ver)):
+            self._failureString += _("Conflicts with the installed package '%s'\n") % cand.name
             return True
         return False
 
@@ -155,12 +175,14 @@ class DebPackage(object):
 
             # check conflicts with virtual pkgs
             if not self._cache.has_key(depname):
+                # FIXME: we have to check for virtual replaces here as 
+                #        well (to pass tests/gdebi-test8.deb)
                 if self._cache.isVirtualPkg(depname):
                     for pkg in self._cache.getProvidersForVirtual(depname):
-                        #print "conflicts virtual check: %s" % pkg.name
+                        self._dbg(3, "conflicts virtual check: %s" % pkg.name)
                         # P/C/R on virtal pkg, e.g. ftpd
                         if self.pkgName == pkg.name:
-                            #print "conflict on self, ignoring"
+                            self._dbg(3, "conflict on self, ignoring")
                             continue
                         if self._checkSinglePkgConflict(pkg.name,ver,oper):
                             self._installedConflicts.add(pkg.name)
@@ -179,11 +201,45 @@ class DebPackage(object):
     def getDepends(self):
         depends = []
         # find depends
-        for key in ["Depends","PreDepends"]:
+        for key in ["Depends","Pre-Depends"]:
             if self._sections.has_key(key):
                 depends.extend(apt_pkg.ParseDepends(self._sections[key]))
         return depends
 
+    def getProvides(self):
+        provides = []
+        key = "Provides"
+        if self._sections.has_key(key):
+            provides = apt_pkg.ParseDepends(self._sections[key])
+        return provides
+
+    def getReplaces(self):
+        replaces = []
+        key = "Replaces"
+        if self._sections.has_key(key):
+            replaces = apt_pkg.ParseDepends(self._sections[key])
+        return replaces
+
+    def replacesRealPkg(self, pkgname, oper, ver):
+        """ 
+        return True if the deb packages replaces a real (not virtual)
+        packages named pkgname, oper, ver 
+        """
+        self._dbg(3, "replacesPkg() %s %s %s" % (pkgname,oper,ver))
+        pkgver = None
+        cand = self._cache[pkgname]
+        if cand.isInstalled:
+            pkgver = cand.installedVersion
+        elif cand.markedInstall:
+            pkgver = cand.candidateVersion
+        for or_group in self.getReplaces():
+            for (name, ver, oper) in or_group:
+                if (name == pkgname and 
+                    apt_pkg.CheckDep(pkgver,oper,ver)):
+                    self._dbg(3, "we have a replaces in our package for the conflict against '%s'" % (pkgname))
+                    return True
+        return False
+
     def checkConflicts(self):
         """ check if the pkg conflicts with a existing or to be installed
             package. Return True if the pkg is ok """
@@ -195,6 +251,57 @@ class DebPackage(object):
                 res = False
         return res
 
+    def checkBreaksExistingPackages(self):
+        """ 
+        check if installing the package would break exsisting 
+        package on the system, e.g. system has:
+        smc depends on smc-data (= 1.4)
+        and user tries to installs smc-data 1.6
+        """
+        # show progress information as this step may take some time
+        size = float(len(self._cache))
+        steps = int(size/50)
+        debver = self._sections["Version"]
+        for (i, pkg) in enumerate(self._cache):
+            if i%steps == 0:
+                self._cache.op_progress.update(float(i)/size*100.0)
+            if not pkg.isInstalled:
+                continue
+            # check if the exising dependencies are still satisfied
+            # with the package
+            ver = pkg._pkg.CurrentVer
+            for dep_or in pkg.installedDependencies:
+                for dep in dep_or.or_dependencies:
+                    if dep.name == self.pkgName:
+                        if not apt_pkg.CheckDep(debver,dep.relation,dep.version):
+                            self._dbg(2, "would break (depends) %s" % pkg.name)
+                            # TRANSLATORS: the first '%s' is the package that breaks, the second the dependency that makes it break, the third the relation (e.g. >=) and the latest the version for the releation
+                            self._failureString += _("Breaks existing package '%(pkgname)s' dependency %(depname)s (%(deprelation)s %i(depversion)s)") % {
+                                'pkgname' : pkg.name, 
+                                'depname' : dep.name, 
+                                'deprelation' : dep.relation, 
+                                'depversion' : dep.version}
+                            self._cache.op_progress.done()
+                            return False
+            # now check if there are conflicts against this package on
+            # the existing system
+            if ver.DependsList.has_key("Conflicts"):
+                for conflictsVerList in ver.DependsList["Conflicts"]:
+                    for cOr in conflictsVerList:
+                        if cOr.TargetPkg.Name == self.pkgName:
+                            if apt_pkg.CheckDep(debver, cOr.CompType, cOr.TargetVer):
+                                self._dbg(2, "would break (conflicts) %s" % pkg.name)
+				# TRANSLATORS: the first '%s' is the package that conflicts, the second the packagename that it conflicts with (so the name of the deb the user tries to install), the third is the relation (e.g. >=) and the last is the version for the relation
+                                self._failureString += _("Breaks existing package '%(pkgname)s' conflict: %(targetpkg)s (%(comptype)s %(targetver)s)") % {
+                                    'pkgname' : pkg.name, 
+                                    'targetpkg' : cOr.TargetPkg.Name, 
+                                    'comptype' : cOr.CompType, 
+                                    'targetver' : cOr.TargetVer }
+                                self._cache.op_progress.done()
+                                return False
+        self._cache.op_progress.done()
+        return True
+
     # some constants
     (NO_VERSION,
      VERSION_OUTDATED,
@@ -230,10 +337,14 @@ class DebPackage(object):
         self._dbg(3,"checkDepends")
 
         # check arch
+        if not self._sections.has_key("Architecture"):
+            self._dbg(1, "ERROR: no architecture field")
+            self._failureString = _("No Architecture field in the package")
+            return False
         arch = self._sections["Architecture"]
         if  arch != "all" and arch != apt_pkg.Config.Find("APT::Architecture"):
             self._dbg(1,"ERROR: Wrong architecture dude!")
-            self._failureString = _("Wrong architecture '%s'" % arch)
+            self._failureString = _("Wrong architecture '%s'") % arch
             return False
 
         # check version
@@ -249,6 +360,14 @@ class DebPackage(object):
         if not self.checkConflicts():
             return False
 
+        # set progress information
+        self._cache.op_progress.subOp = _("Analysing dependencies")
+
+        # check if installing it would break anything on the 
+        # current system
+        if not self.checkBreaksExistingPackages():
+            return False
+
         # try to satisfy the dependencies
         res = self._satisfyDepends(self.getDepends())
         if not res:
@@ -287,7 +406,7 @@ class DebPackage(object):
                 try:
                     self._cache[pkg].markInstall(fromUser=False)
                 except SystemError, e:
-                    self._failureString = _("Cannot install '%s'" % pkg)
+                    self._failureString = _("Cannot install '%s'") % pkg
                     self._cache.clear()
                     return False
         return True
@@ -321,20 +440,85 @@ class DebPackage(object):
         return (install,remove, unauthenticated)
     requiredChanges = property(requiredChanges)
 
+    def control_filelist(self):
+        """ return the list of files in control.tar.gt """
+        content = []
+        for name in DebFile(self.file).control:
+            if name and name != ".":
+                content.append(name)
+        return sorted(content)
+    control_filelist = property(control_filelist)
+
+    def to_hex(self, in_data):
+        hex = ""
+        for (i, c) in enumerate(in_data):
+            if i%80 == 0:
+                hex += "\n"
+            hex += "%2.2x " % ord(c)
+        return hex
+
+    def to_strish(self, in_data):
+        s = ""
+        for c in in_data:
+            if ord(c) < 10 or ord(c) > 127:
+                s += " "
+            else:
+                s += c
+        return s
+        
+    def _get_content(self, part, name, auto_decompress=True, auto_hex=True):
+        data = part.get_content(name)
+        # check for zip content
+        if name.endswith(".gz") and auto_decompress:
+            io = StringIO(data)
+            gz = gzip.GzipFile(fileobj=io)
+            data = _("Automatically decompressed:\n\n")
+            data += gz.read()
+        # auto-convert to hex
+        try:
+            data = unicode(data, "utf-8")
+        except Exception, e:
+            new_data = _("Automatically converted to printable ascii:\n")
+            new_data += self.to_strish(data)
+            return new_data
+        return data
+
+    def control_content(self, name):
+        """ return the content of a specific control.tar.gz file """
+        control = DebFile(self.file).control
+        if name in control:
+            return self._get_content(control, name)
+        return ""
+
+    def data_content(self, name):
+        """ return the content of a specific control.tar.gz file """
+        data = DebFile(self.file).data
+        if name in data:
+            return self._get_content(data, name)
+        return ""
+
     def filelist(self):
         """ return the list of files in the deb """
         files = []
         def extract_cb(What,Name,Link,Mode,UID,GID,Size,MTime,Major,Minor):
             #print "%s '%s','%s',%u,%u,%u,%u,%u,%u,%u"\
             #      % (What,Name,Link,Mode,UID,GID,Size, MTime, Major, Minor)
-            files.append(Name)
+            if Name != "./":
+                files.append(Name)
         try:
-            apt_inst.debExtract(open(self.file), extract_cb, "data.tar.gz")
-        except SystemError, e:
             try:
-                apt_inst.debExtract(open(self.file), extract_cb, "data.tar.bz2")
+                apt_inst.debExtract(open(self.file), extract_cb, "data.tar.gz")
             except SystemError, e:
-                return [_("List of files could not be read, please report this as a bug")]
+                try:
+                    apt_inst.debExtract(open(self.file), extract_cb, "data.tar.bz2")
+                except SystemError, e:
+                    try:
+                        apt_inst.debExtract(open(self.file), extract_cb, "data.tar.lzma")
+                    except SystemError, e:
+                        return [_("List of files could not be read, please report this as a bug")]
+        # IOError may happen because of gvfs madness (LP: #211822)
+        except IOError, e:
+            return [_("IOError during filelist read: %s") % e]
         return files
     filelist = property(filelist)
     
@@ -343,7 +527,7 @@ class DebPackage(object):
         if not self._sections.has_key(item):
             # Translators: it's for missing entries in the deb package,
             # e.g. a missing "Maintainer" field
-            return _("%s is not available" % item)
+            return _("%s is not available") % item
         return self._sections[item]
 
     def _dbg(self, level, msg):
@@ -372,3 +556,5 @@ if __name__ == "__main__":
     print "missing deps: %s" % d.missingDeps
     print d.requiredChanges
 
+    print d.filelist
+    print d.data_content("usr/share/locale/bn/LC_MESSAGES/gdebi.mo")
